/*
 * Implement gettime_us() for Unix.
 */

#include <time.h>
#include <sys/time.h>

#include "putty.h"
#include "tree234.h"
#include "xtruss.h"

struct time_us gettime_us(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    struct time_us toret;
    toret.seconds = tv.tv_sec;
    toret.micros = tv.tv_usec;
    return toret;
}
