/*
 * Common pieces between the platform console frontend modules.
 */

extern const char hk_absentmsg_common_fmt[];
extern const char hk_absentmsg_interactive_intro[];
extern const char hk_absentmsg_interactive_prompt[];
extern const char hk_wrongmsg_common_fmt[];
extern const char hk_wrongmsg_interactive_intro[];
extern const char hk_wrongmsg_interactive_prompt[];

extern const char weakcrypto_msg_common_fmt[];

extern const char weakhk_msg_common_fmt[];

extern const char console_continue_prompt[];
extern const char console_abandoned_msg[];
